@@ -1,5 +1,5 @@
 /*****************************************************************************\
- *  $Id: genders.c 1035 2006-03-22 20:36:00Z grondo $
+ *  $Id: genders.c 1084 2006-12-05 00:09:07Z grondo $
  *****************************************************************************
  *  Copyright (C) 2001-2006 The Regents of the University of California.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
@@ -156,7 +156,6 @@ static hostlist_t _genders_to_altnames(genders_t g, hostlist_t hl);
 static hostlist_t _read_genders(List l);
 static void       _genders_opt_verify(opt_t *opt);
 static int        _delete_all (hostlist_t hl, hostlist_t dl);
-static List       _attrlist_append (List l, char *str);
 static int        register_genders_rcmd_types (opt_t *opt);
 
 
@@ -170,7 +169,7 @@ genders_process_opt(opt_t *pdsh_opts, int opt, char *arg)
 #if !GENDERS_G_ONLY
     case 'a':  
         /* For -a, exclude nodes with "pdsh_all_skip" */ 
-        excllist = _attrlist_append (excllist, "pdsh_all_skip");
+        excllist = list_split_append (excllist, ",", "pdsh_all_skip");
     case 'A':
         allnodes = true;
         break;
@@ -179,10 +178,10 @@ genders_process_opt(opt_t *pdsh_opts, int opt, char *arg)
         break;
 #endif /* !GENDERS_G_ONLY */
     case 'g':
-        attrlist = _attrlist_append (attrlist, arg);
+        attrlist = list_split_append (attrlist, ",", arg);
         break;
     case 'X':
-        excllist = _attrlist_append (excllist, arg);
+        excllist = list_split_append (excllist, ",", arg);
         break;
     case 'F':
         gfile = Strdup (arg);
@@ -515,12 +514,37 @@ attrval_by_altname (genders_t g, const char *host, const char *attr,
     return rc;
 }
 
+/*
+ *  Parse the value of "pdsh_rcmd_type" and split into user and rcmd
+ *   strings, passing rcmd name (if any) in *rp, and user name (if any)
+ *   in *up.
+ *   
+ *  Allows pdsh_rcmd_type to be set to [user@][rcmd], where user@ and 
+ *   rcmd are both optional. (i.e. you can set user or rcmd or both)
+ */
+static int rcmd_type_parse (char *val, char **rp, char **up)
+{
+    char *p;
+    *up = NULL;
+    *rp = NULL;
 
+    if ((p = strchr (val, '@'))) {
+        *(p)++ = '\0';
+        *up = val;
+        if (strlen (p) != 0)
+            *rp = p;
+    } else
+        *rp = val;
+
+    return (0);
+}
 
 static int
 register_genders_rcmd_types (opt_t *opt)
 {
     char *host;
+    char *rcmd;
+    char *user;
     char val[64];
     char rcmd_attr[] = "pdsh_rcmd_type";
     hostlist_iterator_t i = NULL;
@@ -546,9 +570,11 @@ register_genders_rcmd_types (opt_t *opt)
          */
         if (rc < 0 && (genders_errnum(gh) == GENDERS_ERR_NOTFOUND)) 
             rc = attrval_by_altname (gh, host, rcmd_attr, val, sizeof (val));
+        
+        rcmd_type_parse (val, &rcmd, &user);
 
         if (rc > 0) 
-            rcmd_register_defaults (host, val, NULL);
+            rcmd_register_defaults (host, rcmd, user);
 
         free (host);
     }
@@ -573,27 +599,6 @@ _delete_all (hostlist_t hl, hostlist_t dl)
     return (rc);
 }
 
-/*
- *  Split comma-separated "attrs" from str and append to list `lp'
- */
-static List _attrlist_append (List l, char *str)
-{
-    List tmp = list_split (",", str);
-    ListIterator i = NULL;
-    char *attr = NULL;
-
-    if (l == NULL) 
-        return ((l = tmp));
-
-    i = list_iterator_create (tmp);
-    while ((attr = list_next (i))) 
-        list_append (l, Strdup(attr));
-    list_destroy (tmp);
-
-    return (l);
-}
-
-
 /*
  * vi: tabstop=4 shiftwidth=4 expandtab
  */
