@@ -1,5 +1,5 @@
 /*****************************************************************************\
- *  $Id: sshcmd.c 1035 2006-03-22 20:36:00Z grondo $
+ *  $Id: sshcmd.c 1118 2006-12-13 00:03:10Z grondo $
  *****************************************************************************
  *  Copyright (C) 2001-2006 The Regents of the University of California.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
@@ -378,8 +378,9 @@ static int sshcmd_args_init (void)
     Free ((void **) &str);
 
     if ((val = getenv ("PDSH_SSH_ARGS_APPEND"))) {
+        List l;
         str = Strdup (val);
-        List l = list_split (" ", str);
+        l = list_split (" ", str);
         Free ((void **) &str);
 
         while ((str = list_pop (l)))
