@@ -1,5 +1,5 @@
 /*****************************************************************************\
- *  $Id: dsh.c 1035 2006-03-22 20:36:00Z grondo $
+ *  $Id: dsh.c 1123 2007-05-07 16:44:53Z grondo $
  *****************************************************************************
  *  Copyright (C) 2001-2006 The Regents of the University of California.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
@@ -67,9 +67,6 @@
 #include <sys/wait.h>
 #include <sys/poll.h>
 #include <sys/time.h>
-#include <sys/stat.h>
-#include <sys/param.h>
-#include <dirent.h>
 #if	HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -87,21 +84,26 @@
 #include <netdb.h>              /* gethostbyname */
 #include <sys/resource.h>       /* get/setrlimit */
 
-#ifndef MAXPATHNAMELEN
-#define MAXPATHNAMELEN MAXPATHLEN
-#endif
-
 #ifndef PTHREAD_STACK_MIN
 #  define PTHREAD_STACK_MIN ((size_t) sysconf (_SC_THREAD_STACK_MIN))
 #endif
 
+#define dsh_mutex_lock(pmutex)                                                \
+ do {                                                                         \
+      if ((errno = pthread_mutex_lock (pmutex)))                              \
+           errx ("%s:%d: mutex_lock: %m", __FILE__, __LINE__);                \
+  } while (0)
+
+#define dsh_mutex_unlock(pmutex)                                              \
+ do {                                                                         \
+      if ((errno = pthread_mutex_unlock (pmutex)))                            \
+           errx ("%s:%d: mutex_unlock: %m", __FILE__, __LINE__);              \
+  } while (0)
+
+
 /* set the default stacksize for threads to 128k */
 #define DSH_THREAD_STACKSIZE    128*1024
 
-/* define the filename flag as an impossible filename */
-#define EXIT_SUBDIR_FILENAME    "a!b@c#d$"
-#define EXIT_SUBDIR_FLAG        "E\n"
-
 #include "src/common/list.h"
 #include "src/common/xmalloc.h"
 #include "src/common/xstring.h"
@@ -110,6 +112,8 @@
 #include "src/common/fd.h"
 #include "dsh.h"
 #include "opt.h"
+#include "pcp_client.h"
+#include "pcp_server.h"
 #include "wcoll.h"
 #include "rcmd.h"
 
@@ -132,17 +136,25 @@ static int threadcount = 0;
  * report which hosts are blocked.
  */
 static thd_t *t;
+static pthread_mutex_t thd_mutex = PTHREAD_MUTEX_INITIALIZER;
 
 /*
  * Timeout values, initialized in dsh(), used in _wdog().
  */
 static int connect_timeout, command_timeout;
 
+/*
+ * Terminate on a single SIGINT (batch mode)
+ */
+static int sigint_terminates = 0;
+
 /*
  *  Buffered output prototypes:
  */
 typedef void (* out_f) (const char *, ...);
 static int _do_output (int fd, cbuf_t cb, out_f outf, bool read_rc, thd_t *t);
+static int _handle_rcmd_stderr (thd_t *t);
+static int _handle_rcmd_stdout (thd_t *t);
 static void _flush_output (cbuf_t cb, out_f outf, thd_t *t);
 
 /*
@@ -174,7 +186,7 @@ static void _alarm_handler(int dummy)
 }
 
 /* 
- * Helper function for intr_handler().  Lists the status of all connected
+ * Helper function for handle_sigint().  Lists the status of all connected
  * threads.
  */
 static void _list_slowthreads(void)
@@ -182,6 +194,8 @@ static void _list_slowthreads(void)
     int i;
     time_t ttl;
 
+    dsh_mutex_lock(&thd_mutex);
+
     for (i = 0; t[i].host != NULL; i++) {
 
         switch (t[i].state) {
@@ -213,20 +227,30 @@ static void _list_slowthreads(void)
             if (debug)
                 err("%p: %S: [done]\n", t[i].host);
             break;
+        case DSH_CANCELED:
+            if (debug)
+                err("%p: %S: [canceled]\n", t[i].host);
+            break;
         }
     }
+
+    dsh_mutex_unlock(&thd_mutex);
 }
 
 /*
- * Block SIGINT in this thread.
+ * Block SIGINT SIGTSTP SIGCHLD n this thread.
  */
-static void _int_block(void)
+static void _mask_signals(int how)
 {
     sigset_t blockme;
 
+    assert ((how == SIG_BLOCK) || (how == SIG_UNBLOCK));
+
     sigemptyset(&blockme);
     sigaddset(&blockme, SIGINT);
-    pthread_sigmask(SIG_BLOCK, &blockme, NULL);
+    sigaddset(&blockme, SIGTSTP);
+    sigaddset(&blockme, SIGCHLD);
+    pthread_sigmask(how, &blockme, NULL);
 }
 
 /*
@@ -237,44 +261,31 @@ static void _fwd_signal(int signum)
 {
     int i;
 
+    dsh_mutex_lock(&thd_mutex);
     for (i = 0; t[i].host != NULL; i++) {
         if ((t[i].state == DSH_READING)) 
 			rcmd_signal(t[i].rcmd, signum);
     }
+    dsh_mutex_unlock(&thd_mutex);
 
 }
 
-/* 
- * SIGINT handler.  The program can be terminated by two ^C's within
- * INTR_TIME seconds.  Otherwise, ^C causes a list of connected thread
- * status.  This should only be handled by the "main" thread.  We block
- * SIGINT in other threads.
- */
-static void _int_handler(int signum)
+static int _thd_connect_timeout (thd_t *t)
 {
-    static time_t last_intr = 0;
-
-    if (!t) return;
-
-    if (time(NULL) - last_intr > INTR_TIME) {
-        err("%p: interrupt (one more within %d sec to abort)\n",
-            INTR_TIME);
-        last_intr = time(NULL);
-        _list_slowthreads();
-    } else {
-        _fwd_signal(signum);
-        errx("%p: interrupt, aborting.\n");
+    if ((connect_timeout > 0) && (t->start != ((time_t) -1))) {
+        if (t->start + connect_timeout < time (NULL))
+            return (1);
     }
+    return (0);
 }
 
-/*
- * Simpler version of above for -b "batch mode", i.e. pdsh is run by a
- * script, and when the script dies, we should die too.
- */
-static void _int_handler_justdie(int signum)
+static int _thd_command_timeout (thd_t *t)
 {
-    _fwd_signal(signum);
-    errx("%p: batch mode interrupt, aborting.\n");
+    if ((command_timeout > 0) && (t->connect != ((time_t) -1))) {
+        if (t->connect + command_timeout < time (NULL))
+            return (1);
+    }
+    return (0);
 }
 
 /* 
@@ -287,8 +298,6 @@ static void *_wdog(void *args)
 {
     int i;
     
-    _int_block();               /* block SIGINT */
-
     for (;;) {
 
         if (t == NULL) /* We're done */
@@ -297,20 +306,17 @@ static void *_wdog(void *args)
         for (i = 0; t[i].host != NULL; i++) {
             switch (t[i].state) {
             case DSH_RCMD:
-                if ((connect_timeout > 0) && (t[i].start != ((time_t) -1))) {
-                    if (t[i].start + connect_timeout < time(NULL))
+                if (_thd_connect_timeout (&t[i]))
                         pthread_kill(t[i].thread, SIGALRM);
-                }
                 break;
             case DSH_READING:
-                if ((command_timeout > 0) && (t[i].connect != ((time_t) -1))) {
-                    if (t[i].connect + command_timeout < time(NULL))
+                if (_thd_command_timeout (&t[i]))
                         pthread_kill(t[i].thread, SIGALRM);
-                }
                 break;
             case DSH_NEW:
             case DSH_DONE:
             case DSH_FAILED:
+            case DSH_CANCELED:
                 break;
             }
         }
@@ -319,254 +325,6 @@ static void *_wdog(void *args)
     return NULL;
 }
 
-static void _rexpand_dir(List list, char *name)
-{
-    DIR *dir;
-    struct dirent *dp;
-    struct stat sb;
-    char file[MAXPATHNAMELEN];
-
-    dir = opendir(name);
-    if (dir == NULL)
-        errx("%p: opendir: %s: %m\n", name);
-    while ((dp = readdir(dir))) {
-        if (dp->d_ino == 0)
-            continue;
-        if (!strcmp(dp->d_name, ".") || !strcmp(dp->d_name, ".."))
-            continue;
-        snprintf(file, sizeof(file), "%s/%s", name, dp->d_name);
-        if (stat(file, &sb) < 0)
-            errx("%p: can't stat %s: %m\n", file);
-        if (access(name, R_OK) < 0)
-            errx("%p: access: %s: %m\n", name);
-        if (!S_ISDIR(sb.st_mode) && !S_ISREG(sb.st_mode))
-            errx("%p: not a regular file or directory: %s\n", file);
-        list_append(list, Strdup(file));
-        if (S_ISDIR(sb.st_mode))
-            _rexpand_dir(list, file);
-    }
-    closedir(dir);
-
-    /* Since pdcp reads file names and directories only once for
-     * efficiency, we must specify a special flag so we know when
-     * to tell the server to "move up" the directory tree.
-     */
-
-    list_append(list, Strdup(EXIT_SUBDIR_FILENAME));
-}
-
-static List _expand_dirs(List infiles)
-{
-    List new = list_create(NULL);
-    struct stat sb;
-    char *name;
-    ListIterator i;
-
-    i = list_iterator_create(infiles);
-    while ((name = list_next(i))) {
-        if (access(name, R_OK) < 0)
-            errx("%p: access: %s: %m\n", name);
-        if (stat(name, &sb) < 0)
-            errx("%p: stat: %s: %m\n", name);
-        list_append(new, name);
-        /* -r option checked during command line argument checks */
-        if (S_ISDIR(sb.st_mode))
-            _rexpand_dir(new, name);
-    }
-    
-    return new;
-}
-
-/*
- * Wrapper for the write system call that handles short writes.
- * Not sure if write ever returns short in practice but we have to be sure.
- *	fd (IN)		file descriptor to write to 
- *	buf (IN)	data to write
- *	size (IN)	size of buf
- *	RETURN		-1 on failure, size on success
- */
-static int _rcp_write(int fd, char *buf, int size)
-{
-    char *bufp = buf;
-    int towrite = size;
-    int outbytes;
-
-    while (towrite > 0) {
-        outbytes = write(fd, bufp, towrite);
-        if (outbytes <= 0) {
-            assert(outbytes != 0);
-            return -1;
-        }
-        towrite -= outbytes;
-        bufp += outbytes;
-    }
-    return size;
-}
-
-/*
- * Write the contents of the named file to the specified file descriptor.
- *	outfd (IN)	file descriptor to write to 
- *	filename (IN)	name of file
- *	host (IN)	name of remote host for error messages
- *	RETURN		-1 on failure, 0 on success.
- */
-static int _rcp_send_file_data(int outfd, char *filename, char *host)
-{
-    int infd, inbytes, total = 0;
-    char tmpbuf[BUFSIZ];
-
-    infd = open(filename, O_RDONLY);
-    /* checked ahead of time - shouldn't happen */
-    if (infd < 0) {
-        err("%S: _rcp_send_file_data: open %s: %m\n", host, filename);
-        return -1;
-    }
-    do {
-        inbytes = read(infd, tmpbuf, BUFSIZ);
-        if (inbytes < 0) {
-            err("%S: _rcp_send_file_data: read %s: %m\n", host, filename);
-            return -1;
-        }
-        if (inbytes > 0) {
-            total += inbytes;
-            if (_rcp_write(outfd, tmpbuf, inbytes) < 0) {
-                err("%S: _rcp_send_file_data: write: %m\n", host);
-                return -1;
-            }
-        }
-    } while (inbytes > 0);      /* until EOF */
-    close(infd);
-    return 0;
-}
-
-/*
- * Send string to the specified file descriptor.  Do not send trailing '\0'
- * as RCP terminates strings with newlines.
- *	fd (IN)		file descriptor to write to
- *	str (IN)	string to write
- *	host (IN)	name of remote host for error messages
- *	RETURN 		-1 on failure, 0 on success
- */
-static int _rcp_sendstr(int fd, char *str, char *host)
-{
-    int n;
-    assert(strlen(str) > 0);
-    assert(str[strlen(str) - 1] == '\n');
-
-    if ((n = _rcp_write(fd, str, strlen(str))) < 0) 
-        return -1;
-
-    assert(n == strlen(str));
-    return 0;
-}
-
-/*
- * Receive an RCP response code and possibly error message.
- *	fd (IN)		file desciptor to read from
- *	host (IN)	hostname for error messages
- *	RETURN		-1 on fatal error, 0 otherwise
- */
-static int _rcp_response(int fd, char *host)
-{
-    char resp;
-    int i = 0, result = -1;
-    int n;
-    char errstr[BUFSIZ];
-
-    if ((n = read(fd, &resp, sizeof(resp))) != sizeof(resp)) 
-        return (-1);
-
-    switch (resp) {
-        case 0:                /* ok */
-            result = 0;
-            break;
-        default:               /* just error string */
-            errstr[i++] = resp;
-            result = 0;
-        case 1:                /* fatal error + string */
-            fd_read_line (fd, &errstr[i], BUFSIZ - i);
-            err("%p: %S: %s: %s", host, result ? "fatal" : "error", errstr);
-            break;
-    }
-    return result;
-}
-
-#define RCP_MODEMASK (S_ISUID|S_ISGID|S_ISVTX|S_IRWXU|S_IRWXG|S_IRWXO)
-
-static int _rcp_sendfile(int fd, char *file, char *host, bool popt)
-{
-    int result = 0;
-    char tmpstr[BUFSIZ], *template;
-    struct stat sb;
-
-    /*err("%S: %s\n", host, file); */
-
-    if (stat(file, &sb) < 0) {
-        err("%S: %s: %m\n", host, file);
-        goto fail;
-    }
-
-    if (popt) {
-        /* 
-         * 1: SEND stat time: "T%ld %ld %ld %ld\n" 
-         *    (st_mtime, st_mtime_usec, st_atime, st_atime_usec)
-         */
-        snprintf(tmpstr, sizeof(tmpstr), "T%ld %ld %ld %ld\n",
-                 (long) sb.st_mtime, 0L, sb.st_atime, 0L);
-        if (_rcp_sendstr(fd, tmpstr, host) < 0)
-            goto fail;
-
-        /* 2: RECV response code */
-        if (_rcp_response(fd, host) < 0)
-            goto fail;
-    }
-
-    if (S_ISDIR(sb.st_mode)) {
-        /* 
-         * 3a: SEND directory mode: "D%04o %d %s\n"
-         *     (st_mode & RCP_MODEMASK, 0, name)
-         */
-        snprintf(tmpstr, sizeof(tmpstr), "D%04o %d %s\n",
-                 sb.st_mode & RCP_MODEMASK, 0, xbasename(file));
-        if (_rcp_sendstr(fd, tmpstr, host) < 0)
-            goto fail;
-    } else {
-        /* 
-         * 3b: SEND file mode: "C%04o %lld %s\n" or "C%04o %ld %s\n"
-         *    (st_mode & MODE_MASK, st_size, basename(filename))
-         *    Use second template if sizeof(st_size) > sizeof(long).
-         */
-        template = (sizeof(sb.st_size) > sizeof(long)
-                    ? "C%04o %lld %s\n" : "C%04o %ld %s\n");
-        snprintf(tmpstr, sizeof(tmpstr), template,
-                 sb.st_mode & RCP_MODEMASK, sb.st_size, xbasename(file));
-        if (_rcp_sendstr(fd, tmpstr, host) < 0)
-            goto fail;
-    }
-
-    /* 4: RECV response code */
-    if (_rcp_response(fd, host) < 0)
-        goto fail;
-
-    if (S_ISREG(sb.st_mode)) {
-        /* 5: SEND data */
-        if (_rcp_send_file_data(fd, file, host) < 0)
-            goto fail;
-
-        /* 6: SEND NULL byte */
-        if (_rcp_write(fd, "", 1) < 0)
-            goto fail;
-
-        /* 7: RECV response code */
-        if (_rcp_response(fd, host) < 0)
-            goto fail;
-    }
-
-    result = 1;                 /* indicate success */
-  fail:
-    return result;
-}
-
 /*
  * Return the h_addr of a hostname, exiting if there is a lookup failure.
  *	name (IN)	hostname
@@ -583,6 +341,108 @@ static void _gethost(char *name, char *addr)
     memcpy(addr, hp->h_addr_list[0], IP_ADDR_LEN);
 }
 
+/*
+ *  Update thread state to connecting, unless the thread
+ *   has been canceled, in which case close fds if they are open
+ *   and return DSH_CANCELED.
+ */
+static state_t _update_connect_state (thd_t *a)
+{
+    dsh_mutex_lock(&thd_mutex);
+    a->connect = time(NULL);
+    if (a->state != DSH_CANCELED)
+        a->state = DSH_READING;
+    dsh_mutex_unlock(&thd_mutex);
+
+    if (a->state == DSH_CANCELED) {
+        if (a->rcmd->fd >= 0)
+            close (a->rcmd->fd);
+        if (a->rcmd->efd >= 0)
+            close (a->rcmd->efd);
+    }
+
+    return (a->state);
+}
+
+static int _pcp_server (thd_t *th)
+{
+    struct pcp_server svr[1];
+
+    svr->infd =          th->rcmd->fd;
+    svr->outfd =         svr->infd;
+    svr->preserve =      th->pcp_popt;
+    svr->target_is_dir = th->pcp_yopt;
+    svr->outfile =       th->outfile_name;
+
+    return (pcp_server (svr));
+}
+
+static int _pcp_client (thd_t *th)
+{
+    struct pcp_client pcp[1];
+
+    pcp->infd =       th->rcmd->fd;
+    pcp->outfd =      pcp->infd;
+
+    pcp->preserve =   th->pcp_popt;
+    pcp->pcp_client = th->pcp_Zopt;
+    pcp->host =       th->host;
+    pcp->infiles =    th->pcp_infiles;
+
+    return (pcp_client (pcp));
+}
+
+static int _parallel_copy (thd_t *th)
+{
+    int rv = 0;
+    /* 
+     * Run threaded pcp server or client
+     */
+    if (th->pcp_Popt)
+        rv = _pcp_server (th);
+    else
+        rv = _pcp_client (th);
+
+    if ((!th->pcp_Popt && rv < 0) || (th->pcp_Popt)) {
+        /* 
+         *  Copy any pending stderr to user 
+         *   (ignore errors) 
+         *
+         *  Notes: 
+         *
+         *  If the pcp_client was executed, stderr is unlikely b/c
+         *  the pcp_server does not write to stderr and all
+         *  pdsh/pdcp option parameters have been checked for
+         *  correctness.  However, under a disaster situation,
+         *  pdsh/pdcp core code itself could have an error and
+         *  write to stderr.  This call should only be executed if
+         *  there is an error with the pcp_client.  Since the
+         *  client is responsible for closing/finishing the
+         *  connection the below could spin if an error wasn't
+         *  available for reading.
+         *
+         *  If the pcp_server was executed, stderr output could
+         *  come from the pdsh/pdcp core code.  For example, if a
+         *  file has a permission denied error, it would be caught
+         *  in the options checks and output to stderr.  Even if
+         *  there is no stderr, the below cannot spin b/c the
+         *  pcp_client on the remote node will close the
+         *  connection when the copy (or error output) is
+         *  complete.
+         */
+        while (_handle_rcmd_stderr (th) > 0)
+            ;
+        _flush_output (th->errbuf, (out_f) err, th);
+
+    }
+
+    close(th->rcmd->fd);
+    if (th->dsh_sopt)
+        close(th->rcmd->efd);
+
+    return (rv);
+}
+
 /*
  * Rcp thread.  One per remote connection.
  * Arguments are pointer to thd_t entry defined above.
@@ -592,71 +452,50 @@ static void *_rcp_thread(void *args)
     thd_t *a = (thd_t *) args;
     int result = DSH_DONE;      /* the desired outcome */
     int rc;
-
-    _int_block();               /* block SIGINT */
+    char *rcpycmd = NULL;
 
 #if	HAVE_MTSAFE_GETHOSTBYNAME
     if (a->rcmd->opts->resolve_hosts)
         _gethost(a->host, a->addr);
 #endif
     a->start = time(NULL);
+    dsh_mutex_lock(&thd_mutex);
     a->state = DSH_RCMD;
+    dsh_mutex_unlock(&thd_mutex);
+
+    /* For reverse copy, the host needs to be appended to the end of the command */
+    if (a->pcp_Popt) {
+        xstrcat(&rcpycmd, a->cmd);
+        xstrcat(&rcpycmd, " ");
+        xstrcat(&rcpycmd, a->host);
+    }
 
     rcmd_connect (a->rcmd, a->host, a->addr, a->luser, a->ruser, 
-                  a->cmd, a->nodeid, a->dsh_sopt);
+                  (rcpycmd) ? rcpycmd : a->cmd, a->nodeid, a->dsh_sopt);
+
+    if (rcpycmd)
+        Free((void **) &rcpycmd);
 
     if (a->rcmd->fd == -1)
         result = DSH_FAILED;
-    else {
-        a->connect = time(NULL);
-        a->state = DSH_READING;
+    else if (_update_connect_state(a) != DSH_CANCELED) 
+        _parallel_copy(a);
 
-        /* 0: RECV response code */
-        if (_rcp_response(a->rcmd->fd, a->host) >= 0) {
-            ListIterator i;
-            char *name;
-
-            /* Send the files */
-            i = list_iterator_create(a->pcp_infiles);
-            while ((name = list_next(i))) {
-                if (strcmp(name, EXIT_SUBDIR_FILENAME) == 0) {
-                    if (_rcp_sendstr(a->rcmd->fd, EXIT_SUBDIR_FLAG, a->host) < 0)
-                        errx("%p: failed to send exit subdir flag\n");
-                    if (_rcp_response(a->rcmd->fd, a->host) < 0)
-                        errx("%p: failed to exit subdir properly\n");
-                    continue;
-                }
-                _rcp_sendfile(a->rcmd->fd, name, a->host, a->pcp_popt);
-            }
-            list_iterator_destroy(i);
-        } else {
-            /* 
-             *  Copy any pending stderr to user 
-             *   (ignore errors) 
-             *  XXX: Is this necessary. Will there be any data on stderr?
-             */
-            while (_do_output (a->rcmd->efd, a->errbuf, (out_f) err, false, a) > 0)
-                ;
-            _flush_output (a->errbuf, (out_f) err, a);
-        }
-
-        close(a->rcmd->fd);
-        if (a->dsh_sopt)
-            close(a->rcmd->efd);
-    } 
     /* update status */
+    dsh_mutex_lock(&thd_mutex);
     a->state = result;
     a->finish = time(NULL);
+    dsh_mutex_unlock(&thd_mutex);
 
     rc = rcmd_destroy (a->rcmd);
     if ((a->rc == 0) && (rc > 0))
         a->rc = rc;
 
     /* Signal dsh() so another thread can replace us */
-    pthread_mutex_lock(&threadcount_mutex);
+    dsh_mutex_lock(&threadcount_mutex);
     threadcount--;
     pthread_cond_signal(&threadcount_cond);
-    pthread_mutex_unlock(&threadcount_mutex);
+    dsh_mutex_unlock(&threadcount_mutex);
 
     return NULL;
 }
@@ -692,8 +531,7 @@ static int _do_output (int fd, cbuf_t cb, out_f outf, bool read_rc, thd_t *t)
             return (1);
         err ("%p: %S: read: %m\n", t->host);
         return (-1);
-    } else if (rc == 0) 
-        close (fd);
+    } 
 
     while (cbuf_read_line (cb, buf, 8192, 1) > 0) {
 
@@ -727,7 +565,6 @@ static void _flush_output (cbuf_t cb, out_f outf, thd_t *t)
     return;
 }
 
-
 static int _die_if_signalled (thd_t *t)
 {
     int sig;
@@ -743,6 +580,30 @@ static int _die_if_signalled (thd_t *t)
     return (0);
 }
 
+static int _handle_rcmd_stdout (thd_t *th)
+{
+    int rc = _do_output (th->rcmd->fd, th->outbuf, (out_f) out, true, th);
+
+    if (rc <= 0) {
+        close (th->rcmd->fd);
+        th->rcmd->fd = -1;
+    }
+
+    return (rc);
+}
+
+static int _handle_rcmd_stderr (thd_t *th)
+{
+    int rc = _do_output (th->rcmd->efd, th->errbuf, (out_f) err, false, th);
+
+    if (rc <= 0) {
+        close (th->rcmd->efd);
+        th->rcmd->efd = -1;
+    }
+
+    return (rc);
+}
+
 /*
  * Rsh thread.  One per remote connection.
  * Arguments are pointer to thd_t entry defined above.
@@ -755,8 +616,6 @@ static void *_rsh_thread(void *args)
     struct xpollfd xpfds[2];
     int nfds = 1;
 
-    _int_block();               /* block SIGINT */
-
     a->start = time(NULL);
 
 #if	HAVE_MTSAFE_GETHOSTBYNAME
@@ -766,17 +625,16 @@ static void *_rsh_thread(void *args)
     _xsignal (SIGPIPE, SIG_BLOCK);
 
     /* establish the connection */
+    dsh_mutex_lock(&thd_mutex);
     a->state = DSH_RCMD;
+    dsh_mutex_unlock(&thd_mutex);
 
     rcmd_connect (a->rcmd, a->host, a->addr, a->luser, a->ruser,
                   a->cmd, a->nodeid, a->dsh_sopt);
 
     if (a->rcmd->fd == -1) {
         result = DSH_FAILED;    /* connect failed */
-    } else {
-        /* connected: update status for watchdog thread */
-        a->connect = time(NULL);
-        a->state = DSH_READING;
+    } else if (_update_connect_state(a) != DSH_CANCELED) {
 
         fd_set_nonblocking (a->rcmd->fd);
 
@@ -803,10 +661,13 @@ static void *_rsh_thread(void *args)
             /* poll (possibility for SIGALRM) */
             rv = xpoll(xpfds, nfds, -1);
             if (rv == -1) {
-                if (errno == EINTR)
+                if (errno != EINTR) 
+                    err("%p: %S: xpoll: %m\n", a->host);
+                else if (_thd_command_timeout (a))
                     err("%p: %S: command timeout\n", a->host);
                 else
-                    err("%p: %S: xpoll: %m\n", a->host);
+                    continue; /* interrupted by spurious signal */
+
                 result = DSH_FAILED;
                 rcmd_signal (a->rcmd, SIGTERM);
                 break;
@@ -814,13 +675,13 @@ static void *_rsh_thread(void *args)
 
             /* stdout ready or closed ? */
             if (xpfds[0].revents & (XPOLLREAD|XPOLLERR)) {
-                if (_do_output (a->rcmd->fd, a->outbuf, (out_f) out, true, a) <= 0)
+                if (_handle_rcmd_stdout (a) <= 0)
                     xpfds[0].fd = -1;
             }
 
             /* stderr ready or closed ? */
             if (a->dsh_sopt && xpfds[1].revents & (XPOLLREAD|XPOLLERR)) {
-                if (_do_output (a->rcmd->efd, a->errbuf, (out_f) err, false, a) <= 0)
+                if (_handle_rcmd_stderr (a) <= 0)
                     xpfds[1].fd = -1;
             }
 
@@ -837,8 +698,10 @@ static void *_rsh_thread(void *args)
     }
 
     /* update status */
+    dsh_mutex_lock(&thd_mutex);
     a->state = result;
     a->finish = time(NULL);
+    dsh_mutex_unlock(&thd_mutex);
 
     /* flush any pending output */
     _flush_output (a->outbuf, (out_f) out, a);
@@ -855,10 +718,10 @@ static void *_rsh_thread(void *args)
     }
 
     /* Signal dsh() so another thread can replace us */
-    pthread_mutex_lock(&threadcount_mutex);
+    dsh_mutex_lock(&threadcount_mutex);
     threadcount--;
     pthread_cond_signal(&threadcount_cond);
-    pthread_mutex_unlock(&threadcount_mutex);
+    dsh_mutex_unlock(&threadcount_mutex);
     return NULL;
 }
 
@@ -872,6 +735,7 @@ static void _dump_debug_stats(int rshcount)
     time_t conTot = 0, conMin = TIME_T_YEAR, conMax = 0;
     time_t cmdTot = 0, cmdMin = TIME_T_YEAR, cmdMax = 0;
     int failed = 0;
+    int canceled = 0;
     int n;
 
     for (n = 0; n < rshcount; n++) {
@@ -879,6 +743,10 @@ static void _dump_debug_stats(int rshcount)
             failed++;
             continue;
         }
+        if (t[n].state == DSH_CANCELED) {
+            canceled++;
+            continue;
+        }
         assert(t[n].start && t[n].connect && t[n].finish);
 
         conTot += t[n].connect - t[n].start;
@@ -898,6 +766,8 @@ static void _dump_debug_stats(int rshcount)
         err("Command time:  no sucesses\n");
     }
     err("Failures:      %d\n", failed);
+    if (canceled)
+        err("Canceled:      %d\n", canceled);
 }
 
 /*
@@ -963,7 +833,11 @@ static int _thd_init (thd_t *th, opt_t *opt, List pcp_infiles, int i)
     th->pcp_outfile = opt->outfile_name;
     th->pcp_popt = opt->preserve;
     th->pcp_ropt = opt->recursive;
+    th->pcp_yopt = opt->target_is_directory;
+    th->pcp_Popt = opt->reverse_copy;
+    th->pcp_Zopt = opt->pcp_client;
     th->pcp_progname = opt->progname;
+    th->outfile_name = opt->outfile_name;
     th->kill_on_fail = opt->kill_on_fail;
     th->outbuf = cbuf_create (64, 8192);
     th->errbuf = cbuf_create (64, 8192);
@@ -981,6 +855,100 @@ static int _thd_init (thd_t *th, opt_t *opt, List pcp_infiles, int i)
 
 }
 
+static int 
+_cancel_pending_threads (void)
+{
+    int n = 0;
+    int i;
+
+    if (t == NULL) 
+        return (0);
+
+    dsh_mutex_lock (&threadcount_mutex);
+    for (i = 0; t[i].host != NULL; i++) {
+        if ((t[i].state == DSH_NEW) || (t[i].state == DSH_RCMD)) {
+            t[i].state = DSH_CANCELED;
+            ++n;
+        }
+    }
+    err ("%p: Canceled %d pending threads.\n", n);
+    dsh_mutex_unlock (&threadcount_mutex);
+
+    return (0);
+}
+
+/* 
+ *  Handle SIGNINT from signals thread. One ^C lists slow threads.
+ *   Another ^C within one second aborts the job.
+ */
+static void 
+_handle_sigint(time_t *last_intrp)
+{
+    if (!t) return;
+
+    if (sigint_terminates) {
+        _fwd_signal(SIGINT);
+        errx("%p: batch mode interrupt, aborting.\n");
+        /* NORETURN */
+    } else if (time(NULL) - *last_intrp > INTR_TIME) {
+        err("%p: interrupt (one more within %d sec to abort)\n", INTR_TIME);
+        err("%p:  (^Z within %d sec to cancel pending threads)\n", INTR_TIME);
+        *last_intrp = time(NULL);
+        _list_slowthreads();
+    } else {
+        _fwd_signal(SIGINT);
+        errx("%p: interrupt, aborting.\n");
+    }
+}
+
+/*
+ *  ^Z handler. A ^Z within one second of ^C will cancel any
+ *    threads that have not started or are still connecting.
+ *    Otherwise, ^Z has the default behavior of stopping the process.
+ */
+static void 
+_handle_sigtstp (time_t last_intr)
+{
+    if (!t) 
+        return;
+    if (time (NULL) - last_intr > INTR_TIME) 
+        raise (SIGSTOP);
+    else
+        _cancel_pending_threads ();
+}
+
+static void * 
+_signals_thread (void *arg)
+{
+    sigset_t set;
+    time_t last_intr = 0;
+    int signo;
+    int e;
+
+    sigemptyset (&set);
+    sigaddset (&set, SIGINT);
+    sigaddset (&set, SIGTSTP);
+
+    while (t != NULL) {
+        if ((e = sigwait (&set, &signo)) != 0) {
+            if (e == EINTR) continue;
+            err ("sigwait: %s\n", strerror (e));
+        }
+
+        switch (signo) {
+        case SIGINT:  
+            _handle_sigint (&last_intr); 
+            break;
+        case SIGTSTP: 
+            _handle_sigtstp (last_intr); 
+            break;
+        default:
+            err ("%p: Didn't expect to be here.\n");
+        }
+    }
+    return NULL;
+}
+
 /* 
  * Run command on a list of hosts, keeping 'fanout' number of connections 
  * active concurrently.
@@ -990,11 +958,15 @@ int dsh(opt_t * opt)
     int i, rc = 0;
     int rv, rshcount;
     pthread_t thread_wdog;
+    pthread_t thread_sig;
     pthread_attr_t attr_wdog;
+    pthread_attr_t attr_sig;
     List pcp_infiles = NULL;
     hostlist_iterator_t itr;
-    SigFunc *old_int_handler = NULL;
+    const char *domain = NULL;
+    bool domain_in_label = false;
 
+    _mask_signals (SIG_BLOCK);
 
     /*
      *   Initialize rcmd modules...
@@ -1008,10 +980,9 @@ int dsh(opt_t * opt)
 
     /* install signal handlers */
     _xsignal(SIGALRM, _alarm_handler);
+
     if (opt->sigint_terminates)
-        old_int_handler = _xsignal(SIGINT, _int_handler_justdie);
-    else
-        old_int_handler = _xsignal(SIGINT, _int_handler);
+        sigint_terminates = 1;
 
     rshcount = hostlist_count(opt->wcoll);
 
@@ -1038,11 +1009,14 @@ int dsh(opt_t * opt)
     }
 
     /* build PCP command */
-    if (pdsh_personality() == PCP) {
+    if (pdsh_personality() == PCP && !opt->reverse_copy) {
         char *cmd = NULL;
 
         /* expand directories, if any, and verify access for all files */
-        pcp_infiles = _expand_dirs(opt->infile_names);
+        if (!(pcp_infiles = pcp_expand_dirs(opt->infile_names))) {
+            err("%p: unable to build file copy list\n");
+            exit(1);
+        }
 
         xstrcat(&cmd, opt->path_progname); 
         if (opt->recursive)
@@ -1057,6 +1031,30 @@ int dsh(opt_t * opt)
         opt->cmd = cmd;
     }
 
+    if (pdsh_personality() == PCP && opt->reverse_copy) {
+        char *cmd = NULL;
+        char *filename;
+        ListIterator i;
+
+        xstrcat(&cmd, opt->path_progname); 
+
+        if (opt->recursive)
+            xstrcat(&cmd, " -r");
+        if (opt->preserve)
+            xstrcat(&cmd, " -p");
+        xstrcat(&cmd, " -Z ");               /* invoke pcp client */
+
+        i = list_iterator_create(opt->infile_names);
+        while ((filename = list_next(i))) {
+            xstrcat(&cmd, " ");
+            xstrcat(&cmd, filename);
+        }
+        list_iterator_destroy(i);
+
+        /* The 'host' will be appended to the cmd in _rcp_thread */
+        opt->cmd = cmd;
+    }
+
     /* set debugging flag for this module */
     if (opt->debug)
         debug = 1;
@@ -1068,13 +1066,31 @@ int dsh(opt_t * opt)
         errx("%p: hostlist_iterator_create failed\n");
     i = 0;
     while ((t[i].host = hostlist_next(itr))) {
+        char *d;
+        
         assert(i < rshcount);
+
         _thd_init (&t[i], opt, pcp_infiles, i);
+
+        /*
+         * Require domain names in labels if hosts have 
+         *  different domains
+         */
+        if (!domain_in_label && (d = strchr (t[i].host, '.'))) {
+            if (domain == NULL)
+                domain = d;
+            else if (strcmp (d, domain) != 0)
+                domain_in_label = true;
+        }
+
         i++;
     }
     assert(i == rshcount);
     hostlist_iterator_destroy(itr);
 
+    if (domain_in_label)
+        err_no_strip_domain ();
+
     /* set timeout values for _wdog() */
     connect_timeout = opt->connect_timeout;
     command_timeout = opt->command_timeout;
@@ -1083,15 +1099,32 @@ int dsh(opt_t * opt)
     _dsh_attr_init (&attr_wdog, DSH_THREAD_STACKSIZE);
     rv = pthread_create(&thread_wdog, &attr_wdog, _wdog, (void *) t);
 
+    /* start the signals thread */
+    _dsh_attr_init (&attr_sig, DSH_THREAD_STACKSIZE);
+    rv = pthread_create(&thread_sig, &attr_sig, _signals_thread, (void *) t);
+
     /* start all the other threads (at most 'fanout' active at once) */
     for (i = 0; i < rshcount; i++) {
 
         /* wait until "room" for another thread */
-        pthread_mutex_lock(&threadcount_mutex);
+        dsh_mutex_lock(&threadcount_mutex);
 
         if (opt->fanout == threadcount)
             pthread_cond_wait(&threadcount_cond, &threadcount_mutex);
 
+        /*
+         *  Advance past any canceled threads
+         */
+        while ((t[i].state == DSH_CANCELED) && (i < rshcount))
+            ++i;
+        /*
+         *  Abort if no more threads
+         */
+        if (i >= rshcount) {
+            dsh_mutex_unlock(&threadcount_mutex);
+            break;
+        }
+
         /* create thread */
         _dsh_attr_init (&t[i].attr, DSH_THREAD_STACKSIZE);
 #ifdef 	PTHREAD_SCOPE_SYSTEM
@@ -1108,11 +1141,11 @@ int dsh(opt_t * opt)
         }
         threadcount++;
 
-        pthread_mutex_unlock(&threadcount_mutex);
+        dsh_mutex_unlock(&threadcount_mutex);
     }
 
     /* wait for termination of remaining threads */
-    pthread_mutex_lock(&threadcount_mutex);
+    dsh_mutex_lock(&threadcount_mutex);
     while (threadcount > 0)
         pthread_cond_wait(&threadcount_cond, &threadcount_mutex);
 
@@ -1120,10 +1153,10 @@ int dsh(opt_t * opt)
         _dump_debug_stats(rshcount);
 
     /*
-     * Reinstall old handler for SIGINT since _int_handler
-     *  will segfault once we start freeing thd info structures.
+     * Cancel signals thread and unblock SIGINT/SIGTSTP
      */
-    _xsignal(SIGINT, old_int_handler);
+    pthread_cancel(thread_sig);
+    _mask_signals (SIG_UNBLOCK);
 
     /* if -S, our exit value is the largest of the return codes */
     if (opt->ret_remote_rc) {
