@@ -1,5 +1,5 @@
 /*****************************************************************************\
- *  $Id: opt.c 1036 2006-03-22 20:39:20Z grondo $
+ *  $Id: opt.c 1125 2007-06-04 17:27:56Z grondo $
  *****************************************************************************
  *  Copyright (C) 2001-2006 The Regents of the University of California.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
@@ -60,6 +60,7 @@ static void _show_version(void);
 
 static char *exclude_buf = NULL;
 
+
 #define OPT_USAGE_DSH "\
 Usage: pdsh [-options] command ...\n\
 -S                return largest of remote command return values\n"
@@ -77,6 +78,15 @@ Usage: pdcp [-options] src [src2...] dest\n\
 -p                preserve modification time and modes\n"
 /* undocumented "-y"  target must be directory option */
 /* undocumented "-z"  run pdcp server option */
+/* undocumented "-Z"  run pdcp client option */
+
+#define OPT_USAGE_RPCP "\
+Usage: rpdcp [-options] src [src2...] dir\n\
+-r                recursively copy files\n\
+-p                preserve modification time and modes\n"
+/* undocumented "-y"  target must be directory option */
+/* undocumented "-z"  run pdcp server option */
+/* undocumented "-Z"  run pdcp client option */
 
 #define OPT_USAGE_COMMON "\
 -h                output usage menu and quit\n\
@@ -91,18 +101,19 @@ Usage: pdcp [-options] src [src2...] dest\n\
 -w host,host,...  set target node list on command line\n\
 -x host,host,...  set node exclusion list on command line\n\
 -R name           set rcmd module to name\n\
+-N                disable hostname: labels on output lines\n\
 -L                list info on all loaded modules and exit\n"
 /* undocumented "-T testcase" option */
 /* undocumented "-Q" option */
-
+/* undocumented "-K" option -  keep domain name in output */
 
 #if	HAVE_MAGIC_RSHELL_CLEANUP
 #define DSH_ARGS	"sS"
 #else
-#define DSH_ARGS        "S"
+#define DSH_ARGS    "S"
 #endif
-#define PCP_ARGS	"pryz"
-#define GEN_ARGS	"hLR:t:cqf:w:x:l:u:bI:deVT:Q"
+#define PCP_ARGS	"pryzZ"
+#define GEN_ARGS	"hLNKR:t:cqf:w:x:l:u:bI:deVT:Q"
 
 /*
  *  Pdsh options string (for getopt) -- initialized
@@ -124,6 +135,21 @@ pers_t pdsh_personality(void)
     return personality;
 }
 
+/*
+ *  Remote argv array
+ */
+static char **remote_argv;
+static int    remote_argc;
+
+char **pdsh_remote_argv (void)
+{
+    return remote_argv;
+}
+
+int pdsh_remote_argc (void)
+{
+    return remote_argc;
+}
 
 static void
 _init_pdsh_options()
@@ -260,14 +286,19 @@ void opt_default(opt_t * opt, char *argv0)
 
     opt->progname = xbasename(argv0);
 
+    opt->reverse_copy = false;
+
     if (!strcmp(opt->progname, "pdsh") || !strcmp(opt->progname, "dsh"))
         personality = DSH;
     else if (!strcmp(opt->progname, "pdcp") 
             || !strcmp(opt->progname, "dcp")
             || !strcmp(opt->progname, "pcp") )
         personality = PCP;
-    else
-        errx("%p: program must be named pdsh/dsh/pdcp/dcp/pcp\n");
+    else if (!strcmp(opt->progname, "rpdcp")) {
+        personality = PCP;
+        opt->reverse_copy = true;
+    } else
+        errx("%p: program must be named pdsh/dsh/pdcp/dcp/pcp/rpdcp\n");
 
     if (pdsh_options == NULL)
         _init_pdsh_options();
@@ -325,6 +356,8 @@ void opt_default(opt_t * opt, char *argv0)
     opt->preserve = false;
     opt->pcp_server = false;
     opt->target_is_directory = false;
+    opt->pcp_client = false;
+    opt->pcp_client_host = NULL;
 
     return;
 }
@@ -376,6 +409,7 @@ static void wcoll_append (opt_t *opt, char *hosts);
  */
 void opt_args(opt_t * opt, int argc, char *argv[])
 {
+    int wcoll_count = 0;
     int c;
     extern int optind;
     extern char *optarg;
@@ -386,6 +420,9 @@ void opt_args(opt_t * opt, int argc, char *argv[])
 #endif
     while ((c = getopt(argc, argv, pdsh_options)) != EOF) {
         switch (c) {
+        case 'N':
+            opt->labels = false;
+            break;
         case 'L':
             mod_list_module_info();
             exit(0);
@@ -405,8 +442,15 @@ void opt_args(opt_t * opt, int argc, char *argv[])
         case 'w':              /* target node list */
             if (strcmp(optarg, "-") == 0)
                 opt->wcoll = read_wcoll(NULL, stdin);
-            else
-               wcoll_append(opt, optarg);
+            else { 
+                /* Throw  away contents of WCOLL file if they exist 
+                 */
+                if (!(wcoll_count++) && opt->wcoll) {
+                    hostlist_destroy (opt->wcoll);
+                    opt->wcoll = NULL;
+                }
+                wcoll_append(opt, optarg);
+            }
             break;
         case 'x':              /* exclude node list */
             exclude_buf = Strdup(optarg);
@@ -458,6 +502,9 @@ void opt_args(opt_t * opt, int argc, char *argv[])
         case 'h':              /* display usage message */
             _usage(opt);
             break;
+        case 'K':              /* don't strip host domain in output */
+            err_no_strip_domain (); 
+            break;
         case 'y':
             if (pdsh_personality() == PCP)
                 opt->target_is_directory = true;  /* is target a dir? */
@@ -470,6 +517,12 @@ void opt_args(opt_t * opt, int argc, char *argv[])
             else
                 goto test_module_option;
             break;
+        case 'Z':
+            if (pdsh_personality() == PCP)
+                opt->pcp_client = true;          /* run PDCP client */
+            else
+                goto test_module_option;
+            break;
         default: test_module_option:
             if (mod_process_opt(opt, c, optarg) < 0)
                _usage(opt);
@@ -484,6 +537,15 @@ void opt_args(opt_t * opt, int argc, char *argv[])
     if (rcmd_register_default_rcmd(opt->rcmd_name) < 0)
         exit(1);
 
+
+    /* 
+     *  Save beginning of remote argv in case something needs
+     *   to view the unadulterated version (after shell quoting
+     *   applied, etc.)
+     */
+    remote_argc = argc - optind;
+    remote_argv = argv + optind;
+
     /* DSH: build command */
     if (personality == DSH) {
         for (; optind < argc; optind++) {
@@ -495,10 +557,19 @@ void opt_args(opt_t * opt, int argc, char *argv[])
     } else {
         if (!opt->infile_names)
             opt->infile_names = list_create(NULL);
+        
         for (; optind < argc - 1; optind++)
             list_append(opt->infile_names, argv[optind]);
-        if (optind < argc)
-            xstrcat(&opt->outfile_name, argv[optind]);
+        if (optind < argc) {
+            /* If this is the initial pdcp call, the last argument is
+             * the output file/dir.  If this is the pcp_client, the
+             * last argument is the hostname used for connecting.
+             */
+            if (opt->pcp_client)
+                xstrcat(&opt->pcp_client_host, argv[optind]);
+            else
+                xstrcat(&opt->outfile_name, argv[optind]);
+        }
     }
 
     /* ignore wcoll & -x option if we are running pcp server */
@@ -513,6 +584,39 @@ void opt_args(opt_t * opt, int argc, char *argv[])
     }
 }
 
+/* 
+ * Check if infile_names are legit.
+ */
+static int
+_infile_names_check(opt_t * opt)
+{
+    bool verified = true;
+    ListIterator i;
+    char *name;
+
+    i = list_iterator_create(opt->infile_names);
+    while ((name = list_next(i))) {
+        struct stat sb;
+        if (stat(name, &sb) < 0) {
+            err("%p: can't stat %s\n", name);
+            verified = false;
+            continue;
+        }
+        if (!S_ISREG(sb.st_mode) && !S_ISDIR(sb.st_mode)) {
+            err("%p: not a regular file or directory: %s\n", name);
+            verified = false;
+            break;
+        }
+        if (S_ISDIR(sb.st_mode) && !opt->recursive) {
+            err("%p: use -r to copy directories: %s\n", name);
+            verified = false;
+            break;
+        }
+    }
+    list_iterator_destroy(i);
+    return verified;
+}
+
 /*
  * Trap a few option inconsitencies.
  *	opt (IN)	option struct
@@ -548,7 +652,7 @@ bool opt_verify(opt_t * opt)
         verified = false;
     }
 
-    if (!opt->pcp_server) { 
+    if (!opt->pcp_server && !opt->pcp_client) { 
         /* wcoll is required */
         if (opt->wcoll == NULL || hostlist_count(opt->wcoll) == 0) {
             err("%p: no remote hosts specified\n");
@@ -567,10 +671,7 @@ bool opt_verify(opt_t * opt)
     }
 
     /* PCP: must have source and destination filename(s) */
-    if (personality == PCP && !opt->pcp_server) {
-        ListIterator i;
-        char *name;
-
+    if (personality == PCP && !opt->pcp_server && !opt->pcp_client) {
         if (!opt->outfile_name || list_is_empty(opt->infile_names)) {
             err("%p: pcp requires source and dest filenames\n");
             verified = false;
@@ -581,26 +682,32 @@ bool opt_verify(opt_t * opt)
             verified = false;
         }
 
-        i = list_iterator_create(opt->infile_names);
-        while ((name = list_next(i))) {
-            struct stat sb;
-            if (stat(name, &sb) < 0) {
-                err("%p: can't stat %s\n", name);
+        /* If reverse copy, the infiles need not exist locally */
+        if (!opt->reverse_copy) {
+            if (!_infile_names_check(opt))
+                verified = false;
+        }
+
+        /* If reverse copy, the destination must be a directory */
+        if (opt->reverse_copy) {
+            struct stat statbuf;
+
+            if (stat(opt->outfile_name, &statbuf) < 0) {
+                err("%p: can't stat %s\n", opt->outfile_name);
                 verified = false;
-                continue;
             }
-            if (!S_ISREG(sb.st_mode) && !S_ISDIR(sb.st_mode)) {
-                err("%p: not a regular file or directory: %s\n", name);
+
+            if (!S_ISDIR(statbuf.st_mode)) {
+                err("%p: reverse copy dest must be a directory\n");
                 verified = false;
-                break;
-            }
-            if (S_ISDIR(sb.st_mode) && !opt->recursive) {
-                err("%p: use -r to copy directories: %s\n", name);
-                verified = false;
-                break;
             }
         }
-        list_iterator_destroy(i);
+    }
+
+    /* PCP: server and client sanity check */
+    if (personality == PCP && opt->pcp_server && opt->pcp_client) {
+        err("%p: pcp server and pcp client cannot both be set\n");
+        verified = false;
     }
 
     /* PCP: verify options when -z option specified */
@@ -614,6 +721,41 @@ bool opt_verify(opt_t * opt)
             err("%p: output file must be specified with pcp server\n");
             verified = false;
         }
+
+        if (opt->pcp_client_host) {
+            err("%p: pcp client host should not be specified with pcp server\n");
+            verified = false;
+        }
+
+        if (opt->reverse_copy) {
+            err("%p: reverse copy cannot be specified with pcp server\n");
+            verified = false;
+        }
+    }
+
+    /* PCP: verify options when -Z option specified */
+    if (personality == PCP  && opt->pcp_client) {
+
+        opt->reverse_copy = false;
+
+        if (!opt->infile_names || list_is_empty(opt->infile_names)) {
+            err("%p: list source files required for pcp client\n");
+            verified = false;
+        }
+
+        if (opt->outfile_name) {
+            err("%p: output file should not be specified with pcp client\n");
+            verified = false;
+        }
+
+        if (!opt->pcp_client_host) {
+            err("%p: pcp client host must be specified with pcp client\n");
+            verified = false;
+        }
+
+        /* If reverse copy the infiles should exist locally */
+        if (!_infile_names_check(opt))
+            verified = false;
     }
 
     return verified;
@@ -635,7 +777,6 @@ bool opt_verify(opt_t * opt)
  */
 void opt_list(opt_t * opt)
 {
-    char *infile_str;
     char wcoll_str[1024];
     int n;
 
@@ -647,14 +788,11 @@ void opt_list(opt_t * opt)
         out("Appended to cmd         %s\n", STRORNULL(opt->getstat));
         out("Command:		%s\n", STRORNULL(opt->cmd));
     } else {
+        char infiles [4096];
         out("-- PCP-specific options --\n");
-        infile_str = list_join(", ", opt->infile_names);
-        if (infile_str) {
-            out("Infile(s)		%s\n", infile_str);
-            Free((void **) &infile_str);
-        }
-        out("Outfile			%s\n",
-            STRORNULL(opt->outfile_name));
+        if (list_join (infiles, sizeof (infiles), ", ", opt->infile_names))
+            out("Infile(s)		%s\n", infiles);
+        out("Outfile			%s\n", STRORNULL(opt->outfile_name));
         out("Recursive		%s\n", BOOLSTR(opt->recursive));
         out("Preserve mod time/mode	%s\n", BOOLSTR(opt->preserve));
         out("Full program pathname	%s\n", STRORNULL(opt->path_progname));
@@ -722,27 +860,30 @@ void opt_free(opt_t * opt)
  *  Returns a string of comma separated module names of type `type'
  *  Returns NULL if no modules of this type are loaded.
  */
-static char * _module_list_string(char *type)
+static int _module_list_string(char *type, char *buf, int len)
 {
-    List l      = NULL;
-    char *names = NULL;
+    List l = NULL;
+    int  n = 0;
 
     if (mod_count(type) == 0)
-        return NULL;
+        return (0);
 
     l = mod_get_module_names(type);
-    names = list_join(",", l);
+    n = list_join(buf, len, ",", l);
     list_destroy(l);
 
-    return names;
+    return (n);
 }
 
 static char *_rcmd_module_list(char *buf, int maxlen)
 {
     int len, len2;
-    char *rcmd_list = _module_list_string("rcmd");
+    char rbuf [1024];
+    int n;
 
-    len = snprintf(buf, maxlen, "%s", rcmd_list ? rcmd_list : "(none)");
+    n = _module_list_string ("rcmd", rbuf, sizeof (rbuf));
+
+    len = snprintf(buf, maxlen, "%s", n ? rbuf : "(none)");
     if ((len < 0) || (len >= maxlen)) 
         goto done;
 
@@ -756,11 +897,10 @@ static char *_rcmd_module_list(char *buf, int maxlen)
             len += len2;
     }
 
-   done:
+done:
     if ((len < 0) || (len > maxlen)) 
         snprintf(buf + maxlen - 12, 12, "[truncated]"); 
 
-    Free ((void **) &rcmd_list);
     buf[maxlen - 1] = '\0';
     return buf;
 }
@@ -778,8 +918,10 @@ static void _usage(opt_t * opt)
 #if	HAVE_MAGIC_RSHELL_CLEANUP
         err(OPT_USAGE_STDERR);
 #endif
-    } else                      /* PCP */
+    } else if (!opt->reverse_copy) /* PCP */
         err(OPT_USAGE_PCP);
+    else 
+        err(OPT_USAGE_RPCP);
 
     err(OPT_USAGE_COMMON);
 
@@ -795,13 +937,13 @@ static void _show_version(void)
 {
     char buf[1024];
     extern char *pdsh_version;
-    char *misc_list = _module_list_string("misc");
+    int n;
 
     printf("%s\n", pdsh_version);
-    printf("rcmd modules: %s\n", _rcmd_module_list(buf, 1024));
-    printf("misc modules: %s\n", misc_list ? misc_list : "(none)");
+    printf("rcmd modules: %s\n", _rcmd_module_list(buf, sizeof (buf)));
 
-    Free((void **) &misc_list);
+    n = _module_list_string("misc", buf, sizeof (buf));
+    printf("misc modules: %s\n", n ? buf : "(none)");
 
     exit(0);
 }
